import sys
import os


IDAUSR_DEFAULTS = {
    "darwin": "$HOME/.idapro",
    "linux": "$HOME/.idapro",
    "win": "%APPDATA%/Hex-Rays/IDA Pro"
}


def real_path(path):
    abspath = os.path.abspath(path)
    if os.path.isdir(abspath):
        return abspath


def get_virtualenv_path():
    """
    Return the absolute path of the current virtualenv.  Return None if not in
    a virtualenv -- or if the apparent virtualenv path is nonsensical.
    """
    if hasattr(sys, 'real_prefix'):
        if sys.real_prefix != sys.prefix:
            return real_path(sys.prefix)

    elif hasattr(sys, 'base_prefix'):
        if sys.base_prefix != sys.prefix:
            return real_path(sys.prefix)


def get_platform():
    if sys.platform.startswith("linux"):
        return "linux"
    if sys.platform.startswith("win"):
        return "win"
    if sys.platform.startswith("darwin"):
        return "darwin"

    raise RuntimeError("Surprising platform name.")


def append_ida_usr(new_dir, preserve_default=True):
    if "IDAUSR" in os.environ:
        path = os.pathsep.split(os.environ["IDAUSR"])
    else:
        if preserve_default:
            default = os.path.expandvars(IDAUSR_DEFAULTS[get_platform()])
            path = [default]
        else:
            path = []

    path.append(new_dir)
    os.environ["IDAUSR"] = os.pathsep.join(path)



import argparse

from . import manager


def cmd_update(mgr, opts):
    def print_plan(changes):
        if changes["create"]:
            print("  Updated:")
            for ep in changes["create"]:
                print("    - %s.%s" % (ep.dist, ep.name))

        if changes["delete"]:
            print("  Uninstalled:")
            for ep, path in changes["delete"]:
                print("    - %s.%s" % (ep.dist, ep.name))

    changed = False
    for mtype in ["plugins", "loaders", "procs"]:
        plan = mgr.plan_update(mtype)
        if plan["create"] or plan["delete"]:
            print("%s:" % mtype.capitalize())
            mgr.execute_update(mtype, plan)
            print_plan(plan)
            changed = True

    if not changed:
        print("No changes.")


def cmd_status(mgr, opts):
    def print_plan(changes):
        if changes["active"]:
            print("  Active:")
            for ep in changes["active"]:
                print("    - %s.%s" % (ep.dist, ep.name))

        if changes["create"]:
            print("  Need Update:")
            for ep in changes["create"]:
                print("    - %s.%s" % (ep.dist, ep.name))

        if changes["delete"]:
            print("  Uninstalled:")
            for ep, path in changes["delete"]:
                print("    - %s.%s" % (ep.dist, ep.name))

    for mtype in ["plugins", "loaders", "procs"]:
        plan = mgr.plan_update(mtype)
        if any(v for v in plan.values()):
            print("%s:" % mtype.capitalize())
            print_plan(plan)


def cmd_prefix(mgr, opts):
    print(mgr.user_dir)


def cmd_disable(mgr, opts):
    raise NotImplementedError()


def main():
    ap = argparse.ArgumentParser()
    sps = ap.add_subparsers()

    sp = sps.add_parser("update",
                        help="Update installed IDA modules.")
    sp.set_defaults(func=cmd_update)

    sp = sps.add_parser("status",
                        help="List installed IDA modules.")
    sp.set_defaults(func=cmd_status)

    sp = sps.add_parser("prefix", help="Print the idaenv install prefix.")
    sp.set_defaults(func=cmd_prefix)

    sp = sps.add_parser("disable",
                        help="Disable an IDA module.")
    sp.set_defaults(func=cmd_disable)

    opts = ap.parse_args()

    mgr = manager.VenvPluginManager()

    opts.func(mgr, opts)



from __future__ import print_function

import re
import ast
import os
import hashlib
import pkg_resources

from collections import namedtuple

from .utils import get_virtualenv_path


MODULE_TYPES = {"plugins", "loaders", "procs"}


PLUGIN_TEMPLATE = """
import pkg_resources

def PLUGIN_ENTRY():
    ep = pkg_resources.load_entry_point(%(dist)r, %(group)r, %(name)r)
    return ep()
"""

PROC_TEMPLATE = """
import pkg_resources

__ep = pkg_resources.load_entry_point(%(dist)r, %(group)r, %(name)r)
from __ep import *
"""


LOADER_TEMPLATE = """
import pkg_resources

__ep = pkg_resources.load_entry_point(%(dist)r, %(group)r, %(name)r)
from __ep import *
"""


EntryPointInfo = namedtuple("EntryPointInfo", "dist group name")


class PluginManager(object):
    """
    IDAPython plugin manager.
    """
    template_map = {
        "plugins": PLUGIN_TEMPLATE,
        "procs": PROC_TEMPLATE,
        "loaders": LOADER_TEMPLATE,
    }

    wrapper_rx = r"^[a-zA-Z][a-zA-Z0-9_]*_[0-9a-fA-F]+\.py$"

    def __init__(self, user_dir):
        self.user_dir = user_dir
        self.initialize_user_dir(self.user_dir)

        # XXX: Doing this here is a bit of a hack; pkg_resources needs to be
        # re-initialized whenever a module is installed or removed.
        pkg_resources._initialize_master_working_set()

    def initialize_user_dir(self, user_dir):
        if not user_dir:
            raise ValueError("Invalid user directory.")

        if not os.path.isdir(user_dir):
            os.makedirs(user_dir)

        for module_type in MODULE_TYPES:
            subdir_path = self.wrapper_dir(module_type)
            if not os.path.isdir(subdir_path):
                os.mkdir(subdir_path)

    def update_plugins(self):
        "Update the list of installed IDA plugins and wrapper files."
        # Update module wrappers
        for module_type in MODULE_TYPES:
            plan = self.plan_update(module_type)
            self.execute_update(module_type, plan)

    def plan_update(self, module_type):
        "Plan actions for synchronization."
        if module_type not in MODULE_TYPES:
            raise ValueError("Invalid module type: %r" % module_type)

        # Scan installed modules
        cur_modules = self.find_installed_modules(module_type)

        # Scan installed wrappers
        wrapper_dir = self.wrapper_dir(module_type)
        wrappers = self.find_wrappers(wrapper_dir)
        wrapper_set = set(ep for ep, _ in wrappers)

        # Active modules don't need a change
        active_modules = [ep for ep in cur_modules
                          if ep in wrapper_set]

        # Delete wrappers for uninstalled modules
        to_delete = [(ep, path) for ep, path in wrappers
                     if ep not in cur_modules]

        # Create wrappers for new modules
        to_create = [ep for ep in cur_modules
                     if ep not in wrapper_set]

        return {
            "active": active_modules,
            "delete": to_delete,
            "create": to_create
        }

    def execute_update(self, module_type, changes):
        # Delete uninstalled modules
        for ep, wrapper_path in changes["delete"]:
            os.remove(wrapper_path)

        # Create new wrappers
        for ep in changes["create"]:
            self.write_entry_point_wrapper(module_type, ep)

    def write_entry_point_wrapper(self, module_type, ep_info):
        "Create a wrapper file for IDA."
        wrapper = self.template_map[module_type] % ep_info._asdict()

        dst_path = os.path.join(self.wrapper_dir(module_type),
                                self.wrapper_name(ep_info))
        print("Writing wrapper to %r..." % dst_path)
        with open(dst_path, "w") as wf:
            wf.write(wrapper)

    def entry_point_name(self, module_type):
        return "idapython_" + module_type

    def find_installed_modules(self, module_type):
        entry_point_group = self.entry_point_name(module_type)
        cur_modules = set()
        for ep in pkg_resources.iter_entry_points(entry_point_group):
            cur_modules.add(EntryPointInfo(ep.dist.key,
                                           entry_point_group,
                                           ep.name))
        return cur_modules

    def wrapper_dir(self, module_type):
        "Return path to wrappers for a given module type."
        return os.path.join(self.user_dir, module_type)

    def find_wrappers(self, subdir):
        "Return a list of wrappers in a directory."
        wrappers = []
        for name in os.listdir(subdir):
            path = os.path.join(subdir, name)
            if os.path.isfile(path) and re.match(self.wrapper_rx, name):
                info = self.read_wrapper_info(path)
                if info:
                    wrappers.append((info, path))
        return wrappers

    def read_wrapper_info(self, path):
        """
        Extract the entry point loaded by a wrapper.
        """
        with open(path, "r") as f:
            # Wrappers generated by this manager will be small.
            content = f.read(4096)

        m = re.search(r"load_entry_point(\(.*?\))", content)
        if m:
            ep = ast.literal_eval(m.group(1))
            if (isinstance(ep, tuple) and
                    len(ep) == 3 and
                    all(isinstance(elt, str) for elt in ep)):
                return EntryPointInfo(*ep)

    def wrapper_name(self, ep_info):
        def clean(s):
            s = re.sub(r"(?<=[a-z])[A-Z]",
                       lambda m: "_" + m.group(0).lower(), s)
            s = s.lower()
            s = re.sub(r"[^a-zA-Z0-9_]", "", s)
            s = s.strip("_")
            return s

        dist_part = clean(ep_info.dist[:15])
        name_part = clean(ep_info.name[:15])
        sha_part = hashlib.sha1("_".join(ep_info)).hexdigest()[:6]
        return "%s_%s_%s.py" % (dist_part, name_part, sha_part)


class VenvPluginManager(PluginManager):

    def __init__(self):
        cur_env = get_virtualenv_path()
        if not cur_env:
            raise RuntimeError("Not in virtualenv.")

        ida_path = os.path.join(cur_env, "ida")
        super(VenvPluginManager, self).__init__(ida_path)



"""
Automatic plugin loader.
"""

from .manager import PluginManager


class PluginLoader:
    """
    IDAPython Plugin Loader
    """
    def __init__(self):
        self.manager = PluginManager()

    def install(self):
        "Install IDB open hook and initialize user directory."
        self.manager.install()
        import idaapi

        print("Installing plugin loader...")
        self.initialize_temp_dir()
        self.update_plugin_info()

        idaapi.notify_when(idaapi.NW_OPENIDB, self.idb_open_hook)
        idaapi.notify_when(idaapi.NW_TERMIDA, self.ida_term_hook)

    def idb_open_hook(self, nw=None, old_db=None):
        "Hook called when an IDB is opened."
        # self.update_plugin_info()
        pass

    def ida_term_hook(self, nw_code=None):
        "Hook called when an IDA is terminated."
        pass


# Default global plugin loader instance
_plugin_loader = PluginLoader()


def install_plugin_loader():
    "Watch installed IDA plugins and reload them when opening an IDB."
    global _plugin_loader
    _plugin_loader.install()



from setuptools import setup, find_packages


setup(name='idaenv',
      version="0.1",
      description="IDA Pro plugin manager.",
      packages=find_packages(exclude=['ez_setup']),
      zip_safe=False,
      license="MIT",
      classifiers=[
          "Natural Language :: English",
          "Programming Language :: Python"
      ],
      entry_points={
          "console_scripts": [
              "idaenv=idaenv.command_line:main",
          ]
      })
